/*
 * Copyright (C) 2012  Arnaud Bonatti  <arnaud.bonatti@gmail.com>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

const Clutter		= imports.gi.Clutter
const Gio			= imports.gi.Gio
const Gtk			= imports.gi.Gtk
const GLib			= imports.gi.GLib
const IconGrid		= imports.ui.iconGrid
const Lang			= imports.lang
const Main			= imports.ui.main
const Meta			= imports.gi.Meta
const Params		= imports.misc.params
const RunDialog		= imports.ui.runDialog
const Search		= imports.ui.search
const Shell			= imports.gi.Shell
const St			= imports.gi.St
const Util			= imports.misc.util
const ViewSelector	= imports.ui.viewSelector
const Wanda			= imports.ui.wanda

// for the viewSelector fork
const AppDisplay	= imports.ui.appDisplay
const SearchDisplay	= imports.ui.searchDisplay
const ShellEntry	= imports.ui.shellEntry
const WorkspacesView= imports.ui.workspacesView

// utils
const MainFile		= imports.misc.extensionUtils.getCurrentExtension().imports.extension	// local file
const gettextDomain	= 'terminal-provider'
const _				= imports.gettext.domain(gettextDomain).gettext
const S_			= imports.gettext.domain('gnome-shell').gettext
const MAGIC_FISH_KEY= Wanda.MAGIC_FISH_KEY.toLowerCase()

// schemas
const dconfSchema	= 'org.gnome.shell.extensions.terminal-provider'

const TERM_SCHEMA	= 'org.gnome.desktop.default-applications.terminal'
const TERMNAME_KEY	= 'exec'		// no loading of 'exec-arg': use of "-x" for gnome-terminal and "-e" for others

const LOCK_SCHEMA	= 'org.gnome.desktop.lockdown'
const NO_SHELL_KEY	= 'disable-command-line'

// strings
const EXEC_TEXT		= "Launch command"
const SHELL_TEXT	= "Launch in shell"
const TERM_TEXT		= "Exec in terminal"
const FILE_TEXT		= "Try opening file"
const FOLDER_TEXT	= "Open folder"
const INSTALL_TEXT	= "Install"
const LG_ABREV		= "lg"
const LG_TEXT		= "Looking Glass"
const OVERVIEWTITLE	= "COMMAND LINE"

// icons
const EXEC_ICON		= "system-run"
const TERM_ICON		= "gnome-terminal"
const FOLDER_ICON	= "folder"
const INSTALL_ICON	= "system-software-install"
const FILE_ICON		= "document-open" // or "text-x-generic"? urgh.
const DEVS_ICON		= "applications-development"

let TERMS

/* Remove <Alt>F2, one time */

const SHELLENTRYKEY = 'org.gnome.desktop.wm.keybindings'

let lockFilePath = null
let connectionAltF2 = null
let seSettings = null

function disableShellEntry() {
	seSettings = new Gio.Settings({ 'schema': SHELLENTRYKEY })
	MainFile.getSettings(dconfSchema).set_strv('shell-entry-keybinding', seSettings.get_strv('panel-run-dialog'))
	seSettings.set_strv('panel-run-dialog', [])		// for the control-center not showing the old keybinding

	// not possible to play directly with built-in keybindings (22-08-2012, -> 3.6), so lets play a little
	Meta.keybindings_set_custom_handler('panel-run-dialog', function() {})
	connectionAltF2 = seSettings.connect('changed::panel-run-dialog', function() { if (seSettings.get_strv('panel-run-dialog')!="") rEnableShellEntry(true) })	// TODO test "changed::panel-run-dialog"
}

function rEnableShellEntry(createLockFile) {
	Meta.keybindings_set_custom_handler('panel-run-dialog', Main.openRunDialog)
	seSettings.disconnect(connectionAltF2)
	
	if (createLockFile) {
		let file = Gio.file_new_for_path(lockFilePath)
		let out = file.replace(null, false, Gio.FileCreateFlags.NONE, null)
		Shell.write_string_to_stream (out, "true")
	} else
		seSettings.set_strv('panel-run-dialog', MainFile.getSettings(dconfSchema).get_strv('shell-entry-keybinding'))
	
	connectionAltF2 = null
	seSettings = null
}

/* Terminal Provider */
function TermProvider(completer) { this._init(completer) }

TermProvider.prototype = {
	__proto__: Search.SearchProvider.prototype,

	_init: function(completer) {
		Search.SearchProvider.prototype._init.call(this, _(OVERVIEWTITLE))
		this._completer = completer
		this.words = new Array()
		
		this._connectionTools = global.settings.connect('changed::development-tools', Lang.bind(this, function () {		// as in js/ui/runDialog.js:174
			this._enableLG = global.settings.get_boolean('development-tools')
		}))
		/* this._textChangeConnection = Main.overview._searchEntry.clutter_text.connect('text-changed', Lang.bind(this, function() {
			if (this.pid) GLib.spawn_close_pid(this.pid)
		})) */
		this._enableLG = global.settings.get_boolean('development-tools')
		
		this._appSystem = Shell.AppSystem.get_default()
		
		// TODO need a SIMPLE way to middle-click to open on a new workspace, not as in appDisplay.js:521; instead of using createResultActor (see searchDisplay.js:34), hack
	},

	getResultMetas: function(ids, callback) {
		let textureCache = St.TextureCache.get_default()
		let cb = new Array()
		for (let i in ids) switch (ids[i]) {
			// ONLY ONE TERM
			case "file":
				this.file = Gio.file_new_for_path(this.words[0])
				
				let icon = new Gio.FileIcon({ file: this.file })
				let thumbnail = this.file.query_info(Gio.FILE_ATTRIBUTE_THUMBNAIL_PATH, Gio.FileQueryInfoFlags.NONE, null).get_attribute_as_string(Gio.FILE_ATTRIBUTE_THUMBNAIL_PATH)		// ..._THUMBNAILING_FAILED fails?		// pdf assume having a thumbnail but haven't :(		// TODO what is GnomeDesktopThumbnailFactory?
				if (thumbnail)	cb.push({ 'id': "file",	'name': _(FILE_TEXT),
					'createIcon': function(size) { return textureCache.load_gicon(null, icon, size)} })
				else 			cb.push({ 'id': "file",	'name': _(FILE_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, FILE_ICON, size)} })
				break
			case "folder":		cb.push({ 'id': "folder",	'name': _(FOLDER_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, FOLDER_ICON, size)} })
				break
			case "install":		cb.push({ 'id': "install",	'name': _(INSTALL_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, INSTALL_ICON, size)} })
				break
			
			// ONE TERM OR +
			case "app":
				let word = this.words[0].slice(this.words[0].lastIndexOf('/') + 1).concat(".desktop")
				cb.push({ 'id': "exec",		'name': _(EXEC_TEXT),
					'createIcon': function(size) { return Shell.AppSystem.get_default().lookup_app(word).create_icon_texture(size) }})
				break
			case "exec":		cb.push({ 'id': "exec",		'name': _(EXEC_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, EXEC_ICON, size)} })
				break
			case "shell":		cb.push({ 'id': "shell",	'name': _(SHELL_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, EXEC_ICON, size)} })
				break
			case "files":		cb.push({ 'id': "files",	'name': _(FILE_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, FILE_ICON, size)} })
				break
			case "term":		cb.push({ 'id': "term",		'name': _(TERM_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, TERM_ICON, size)} })
				break
			case "lg":			cb.push({ 'id': "lg",		'name': _(LG_TEXT),
					'createIcon': function(size) { return textureCache.load_icon_name(null, DEVS_ICON, size)} })
				break
			default: throw new Error("Line 181: " + ids[i])
		}
		callback(cb)
	},
	
	getInitialResultSet: function(terms) {
		this.terms = terms
		
		if (this.pid) GLib.spawn_close_pid(this.pid)
		
		if (terms.length == 1) {
			let path = this._completer.absolutePath(terms[0])
			
			if (GLib.file_test(path, GLib.FileTest.IS_DIR)) {
				this.words = [path]
				this.searchSystem.pushResults(this, ["folder"])
			} else if (GLib.file_test(path, GLib.FileTest.IS_REGULAR)) {
				this.words = [path]
				if (GLib.file_test(path, GLib.FileTest.IS_EXECUTABLE)) {
					let word = path.slice(path.lastIndexOf('/') + 1).concat(".desktop")
					if (this._appSystem.lookup_app(word))
						this.searchSystem.pushResults(this, ["app", "term"])
					else
						this.searchSystem.pushResults(this, ["file", "exec", "term"])
				} else 
					this.searchSystem.pushResults(this, ["file"])
			} else {
				this.words = terms
				let text = terms[0].toLowerCase()
				if (this._completer.checkExists(terms[0])) {
					let word = terms[0].slice(terms[0].lastIndexOf('/') + 1).concat(".desktop")
					if (this._appSystem.lookup_app(word))
						this.searchSystem.pushResults(this, ["app", "term"])
					else
						this.searchSystem.pushResults(this, ["exec", "term"])
				} else if (this._enableLG && text==LG_ABREV)
					this.searchSystem.pushResults(this, ["lg", "shell", "term"])
				else if (install != "") {
					if ( ! ( /^[-.,_0-9a-zA-Z]+$/.test(terms[0]) ) ) { this.searchSystem.pushResults(this, []); return }	// command line injection possible so santize
					
					this.searched_app_name = terms[0]
					let argv
					
					if		(install == "apt")	argv = ["apt-cache", "-q=2", "-n", "show", terms[0]]
					else if	(install == "pk")	argv = ["pkcon", "-p", "get-depends", terms[0]]	// 'get-depends' is slow (still sort of acceptable ...), 'resolve' is faster - but doesnt return exit code > 0 on fail; TODO rethink? it's possible now to process its output, see https://bugzilla.gnome.org/show_bug.cgi?id=667069
					else { this.searchSystem.pushResults(this, []); return }
				
					if (this.pid) GLib.spawn_close_pid(this.pid)
					let [success, pid] = GLib.spawn_async(null, argv, null, GLib.SpawnFlags.SEARCH_PATH | GLib.SpawnFlags.DO_NOT_REAP_CHILD, null, null)
					if (!success) { this.searchSystem.pushResults(this, []); return }
					this.pid = pid
				
					let _self = this
					GLib.child_watch_add(GLib.PRIORITY_DEFAULT, pid, function(pid, status) {
						GLib.spawn_close_pid(pid)
						if (status == 0 && _self.terms[0] == terms[0]) _self.searchSystem.pushResults(_self, ["install"])
					})
				} else
					this.searchSystem.pushResults(this, [])
			}
		} else {
			let [words, is_a_path] = this._completer.parseCommandLine(terms, false)
			
			if (words[0] == this._completer.absolutePath(terms.join(' '))) {
				this.words = words
				if (GLib.file_test(words[0], GLib.FileTest.IS_DIR))
					this.searchSystem.pushResults(this, ["folder"])
				else if (GLib.file_test(words[0], GLib.FileTest.IS_REGULAR)) {
					if (GLib.file_test(words[0], GLib.FileTest.IS_EXECUTABLE))
						this.searchSystem.pushResults(this, ["file", "exec", "term"])
					else
						this.searchSystem.pushResults(this, ["file"])
				} else
					this.searchSystem.pushResults(this, [])
			} else if (is_a_path) {
				[words, number_of_paths] = this._completer.parseCommandLine(terms, true)
				this.words = words
				if (number_of_paths == words.length) {
					if (GLib.file_test(words[0], GLib.FileTest.IS_REGULAR) && GLib.file_test(words[0], GLib.FileTest.IS_EXECUTABLE)) {
						let word = words[0].slice(words[0].lastIndexOf('/') + 1).concat(".desktop")
						if (this._appSystem.lookup_app(word))
							this.searchSystem.pushResults(this, ["app", "shell", "term"])
						else
							this.searchSystem.pushResults(this, ["exec", "files", "term"])
					} else
						this.searchSystem.pushResults(this, ["files"])
				} else if (GLib.file_test(words[0], GLib.FileTest.IS_REGULAR) && GLib.file_test(words[0], GLib.FileTest.IS_EXECUTABLE))
					this.searchSystem.pushResults(this, ["shell", "term"])
				else this.searchSystem.pushResults(this, [])		// case "~ Doc"
			} else {
				let text = terms.join(' ').toLowerCase()
				if (text == MAGIC_FISH_KEY)		// "free" exists...
						this.searchSystem.pushResults(this, [])
				else if (this._completer.checkExists(terms[0])) {
					[words, number_of_paths] = this._completer.parseCommandLine(terms, true)
					this.words = words
					
					if (number_of_paths == words.length - 1) {
						let word = this.words[0].slice(this.words[0].lastIndexOf('/') + 1).concat(".desktop")
						if (this._appSystem.lookup_app(word))
							this.searchSystem.pushResults(this, ["app", "shell", "term"])
						else
							this.searchSystem.pushResults(this, ["exec", "shell", "term"])
					} else
						this.searchSystem.pushResults(this, ["shell", "term"])
				} else
					this.searchSystem.pushResults(this, ["term"])
			}
		}
	},
	getSubsearchResultSet: function(previousResults, terms) {
		this.getInitialResultSet(terms)
	},
	activateResult: function(id) {
		if		(id == "exec")		Util.spawn(this.words)
		else if	(id == "shell")		Util.spawnCommandLine(this.terms.join(' '))
		else if (id == "term")		this._execTerm()
		
		else if (id == "folder")	this._openFile(Gio.file_new_for_path(this.words[0]))
		else if	(id == "file")		this._openFile(this.file)		// see line 127 and 291
		else if	(id == "files")		for (let number in this.words) this._openFile(Gio.file_new_for_path(this.words[number]))	// TODO: unknown files?
		
		else if (id == "install")	this._install()
		else /* (id == "lg") */		Main.createLookingGlass().toggle()
	},
	dragActivateResult: function(id, params) {		// same as activateResult but for d'n'd, see searchDisplay.js:95
		params = Params.parse(params, { workspace: -1, timestamp: 0 })		// call on workspaceThumbnail.js:477
		let launchWorkspace = global.screen.get_workspace_by_index(params.workspace)
		launchWorkspace.activate(global.get_current_time())
		
		this.activateResult(id)
		if (id == "lg") Main.overview.hide()		// Overview auto-hide with activateResult() but not with dragActivateResult()
	},
	_install: function() {
		if		(install == "apt")	Util.spawn(["apturl", "apt://" + this.terms[0]])
		else if (install == "pk")	Util.spawn(["gpk-install-package-name", this.terms[0]])
	},
	_openFile: function(file) {
		try { Gio.app_info_launch_default_for_uri(file.get_uri(), global.create_app_launch_context()) }
		catch (e) {
			// The exception from gjs contains an error string like:
			//		Error invoking Gio.app_info_launch_default_for_uri: No application
			//		is registered as handling this file
			// We are only interested in the part after the first colon.
			let message = e.message.replace(/[^:]*: *(.+)/, '$1')
			global.log(message)
		}
	},
	_execTerm: function() {
		let i = 0, error
		do {
			error = null
			try {
				if (TERMS[i+1])	Util.trySpawn([TERMS[i], (TERMS[i]=="gnome-terminal")?"-x":"-e"].concat(this.terms))
				else			Util.spawn([TERMS[i], (TERMS[i]=="gnome-terminal")?"-x":"-e"].concat(this.terms))
			} catch (err) {
				error = err
				global.log(TERMS[i] + " not found (extension306.js:322). Trying " + TERMS[i+1] + ".")
				i += 1
			}
		} while (error)
	},
	byebye: function() {
		// Main.overview._searchEntry.clutter_text.disconnect(this._textChangeConnection)
		global.settings.disconnect(this._connectionTools)
		this._connectionTools = 0
	}
}

/* Autocompletion */

function Completor() { this._init() }
Completor.prototype = {
	__proto__: RunDialog.CommandCompleter.prototype,
	
	_init: function() {
		RunDialog.CommandCompleter.prototype._init.call(this)
		
		this._pathsCompleter = new Gio.FilenameCompleter()
		this._commCompUpdateConnection = Main.overview.connect('shown', Lang.bind(this, this.update))
	},
	byebye: function() {
		Main.overview.disconnect(this._commCompUpdateConnection)
		this._pathsCompleter = null
	},
	
	checkExists: function(text) {
		for (let i in this._childs) if (GLib.file_test(GLib.build_filenamev([this._paths[i], text]), GLib.FileTest.EXISTS)) return true
		return false
	},
	
	absolutePath: function(text) {
		// Glib.file_test("~/blabla", etc.) fails, Glib.file_test("blabla", etc.) is ok
		if (text.charAt(0) == '/') return text
		if (text == '~') return GLib.get_home_dir()
		if (text.slice(0,2) == '~/' || text.slice(0,2) == '~ ') return GLib.build_filenamev([GLib.get_home_dir(), text.slice(1)])
		return GLib.build_filenamev([GLib.get_home_dir(), text])
	},
	
	parseCommandLine: function(array, recursive) {
		let i, j
		let words = new Array()
		let number_of_paths = 0
		
		for (i=0; i<array.length; i++) {
			let text
			if (array[i] != '~') {
				for (j=array.length; j>i; j--) {
					text = this.absolutePath(array.slice(i,j).join(' '))
					if (GLib.file_test(text, GLib.FileTest.EXISTS)) {
						i = j - 1
						number_of_paths += 1
						break	// the j loop
					}
					if (j == i + 1) text = array[i]
				}
			} else {
				text = GLib.get_home_dir()
				number_of_paths += 1
			}
			words.push(text)
			if (!recursive) break
		}
		return [words, number_of_paths]
	},
	
	pathCompletion: function(text) {		// RunDialog.CommandCompleter also checks home path, time to fork?
		let index
		do {
			text = this.absolutePath(text)		// this._pathsCompleter.get_completion_suffix("~/blabla") is ok, but unused here
			if (GLib.file_test(text.slice(0, text.lastIndexOf('/') + 1), GLib.FileTest.IS_DIR)) return this._pathsCompleter.get_completion_suffix(text)
			index = text.indexOf(' ') + 1
			text = text.slice(index)
		} while (index)				// case [abc/def without ~/abc/def] and case [abc def without ~/abc\ def]
		return false				// case [~/def] is tested by _getCompletion, same code as for when there isn't '/' initial text
	},
	getRealCompletion: function(text) {
		let test = this.pathCompletion(text)
		if (test) return test

		let index
		do {
			index = text.indexOf(' ') + 1
			text = text.slice(index)
			test = this.getCompletion(text)		// TODO: fix the "looking glade" autocompletion problem
			if (test) return test
		} while (index)
		return false
	},
	preload: function(text) {
		if (text.indexOf('/') != -1) this.pathCompletion(text.concat('a'))		// GFilenameCompleter (this._pathsCompleter) loads content of parent directory; so add a fake thing ('a') to permit preloading when last character is '/'
		else this._pathsCompleter.get_completion_suffix('~/a')
	}
}

const ShellEntryWithTabKeyOK = new Lang.Class({
	Name: 'ShellEntryWithTabKeyOK',
	Extends: St.Entry,
	
	_init: function(params, completer) {
		// params = Params.parse(params, { hint_text: _("Type command..."), track_hover: false, can_focus: true })
		this.parent(params)

		this._haveToPreload = true
		this._textInsertConnection = this.clutter_text.connect('insert-text', Lang.bind(this, function(clutter_text, chars, is_empty) {
			if (chars.indexOf('/') != -1) this._haveToPreload = true
		}))
		this._textDeleteConnection = this.clutter_text.connect('delete-text', Lang.bind(this, function() {
			this._haveToPreload = true
		}))
		this._cursorEventConnection = this.clutter_text.connect('cursor-event', Lang.bind(this, function(clutter_text, geometry) {
			let _cursorPosition = clutter_text.get_cursor_position()
			// -1 is before the last character for slice, and after for cursor_get_position(), insert_text() and others, both used for completion...
			this._cursorPosition = (_cursorPosition == -1) ? clutter_text.get_text().length : _cursorPosition
			
			if (!this._haveToPreload) return
			completer.preload(clutter_text.get_text().slice(0, this._cursorPosition))
			this._haveToPreload = false
		}))
		this._autoCompletionConnection = this.clutter_text.connect('key-press-event', Lang.bind(this, function(clutter_text, event) {
			if (event.get_key_symbol() == Clutter.Tab) {
				let postfix = completer.getRealCompletion(this.clutter_text.get_text().slice(0, this._cursorPosition))
				if (postfix != null && postfix.length > 0) {
					clutter_text.insert_text(postfix, this._cursorPosition)
					clutter_text.set_cursor_position(this._cursorPosition + postfix.length)
					clutter_text.delete_selection()		// TODO make getRealCompletion really faster to remove that line :)
				}
				return true
			}
			return false
		}))
	},
	
	byebye: function() {
		this.clutter_text.disconnect(this._textInsertConnection)
		this.clutter_text.disconnect(this._textDeleteConnection)
		this.clutter_text.disconnect(this._cursorEventConnection)
		
		this.destroy()
	}
})

const ViewSelectorWithNamedConnections = new Lang.Class({		// Credit to authors of the file js/ui/viewSelector.js, GPL v. 2
	Name: 'ViewSelectorWithNamedConnections',					// The main change is the naming of connections (for disabling).
	Extends: ViewSelector.ViewSelector,							// Providers are not reloaded, they are imported as a 3rd param.
	
	_init: function(searchEntry, showAppsButton, providers) {
	    this.actor = new St.BoxLayout({ name: 'viewSelector',
	                                    vertical: true });

	    this._showAppsButton = showAppsButton;
	    this._showAppsButton.connect('notify::checked', Lang.bind(this, this._onShowAppsButtonToggled));

	    this._pageArea = new Shell.Stack();
	    this.actor.add(this._pageArea, { x_fill: true,
	                                     y_fill: true,
	                                     expand: true });

	    this._activePage = null;

	    this.active = false;
	    this._searchPending = false;
	    this._searchTimeoutId = 0;

	    this._searchSystem = new Search.SearchSystem();

	    this._entry = searchEntry;
	    ShellEntry.addContextMenu(this._entry);

	    this._text = this._entry.clutter_text;
	    this._connectionCt1 = this._text.connect('text-changed', Lang.bind(this, this._onTextChanged));
	    this._connectionCt2 = this._text.connect('key-press-event', Lang.bind(this, this._onKeyPress));
	    this._connectionCt3 = this._text.connect('key-focus-in', Lang.bind(this, function() {
	        this._searchResults.highlightDefault(true);
	    }));
	    this._connectionCt4 = this._text.connect('key-focus-out', Lang.bind(this, function() {
	        this._searchResults.highlightDefault(false);
	    }));
	    this._connectionSe1 = this._entry.connect('notify::mapped', Lang.bind(this, this._onMapped));
	    this._connectionGl1 = global.stage.connect('notify::key-focus', Lang.bind(this, this._onStageKeyFocusChanged));

	    this._inactiveIcon = new St.Icon({ style_class: 'search-entry-icon',
	                                       icon_name: 'edit-find-symbolic' });
	    this._activeIcon = new St.Icon({ style_class: 'search-entry-icon',
	                                     icon_name: 'edit-clear-symbolic' });
	    this._entry.set_secondary_icon(this._inactiveIcon);

	    this._iconClickedId = 0;
	    this._capturedEventId = 0;

	    this._workspacesDisplay = new WorkspacesView.WorkspacesDisplay();
	    this._workspacesPage = this._addPage(this._workspacesDisplay.actor, null,
	                                         _("Windows"), 'text-x-generic-symbolic');

	    this._appDisplay = new AppDisplay.AllAppDisplay();
	    this._appsPage = this._addPage(this._appDisplay.actor, null,
	                                   _("Applications"), 'system-run-symbolic');

	    this._searchResults = new SearchDisplay.SearchResults(this._searchSystem);
	    this._searchPage = this._addPage(this._searchResults.actor, this._entry,
	                                     _("Search"), 'edit-find-symbolic');

	    // Default search providers
	    for (i in providers) this.addSearchProvider(providers[i])

	    // Since the entry isn't inside the results container we install this
	    // dummy widget as the last results container child so that we can
	    // include the entry in the keynav tab path
	    this._focusTrap = new ViewSelector.FocusTrap({ can_focus: true });
	    this._connectionFc1 = this._focusTrap.connect('key-focus-in', Lang.bind(this, function() {
	        this._entry.grab_key_focus();
	    }));
	    this._searchResults.actor.add_actor(this._focusTrap);

	    global.focus_manager.add_group(this._searchResults.actor);

	    this._connectionOv1 = Main.overview.connect('item-drag-begin',
	                          Lang.bind(this, this._resetShowAppsButton));

	    this._stageKeyPressId = 0;
	    this._connectionOv2 = Main.overview.connect('showing', Lang.bind(this,
	        function () {
	            this._resetShowAppsButton();
	            this._stageKeyPressId = global.stage.connect('key-press-event',
	                                                         Lang.bind(this, this._onStageKeyPress));
	        }));
	    this._connectionOv3 = Main.overview.connect('hiding', Lang.bind(this,
	        function () {
	            this._resetShowAppsButton();
	            if (this._stageKeyPressId != 0) {
	                global.stage.disconnect(this._stageKeyPressId);
	                this._stageKeyPressId = 0;
	            }
	        }));

	    // Public constraints which may be used to tie actors' height or
	    // vertical position to the current tab's content; as the content's
	    // height and position depend on the view selector's style properties
	    // (e.g. font size, padding, spacing, ...) it would be extremely hard
	    // and ugly to get these from the outside. While it would be possible
	    // to use position and height properties directly, outside code would
	    // need to ensure that the content is properly allocated before
	    // accessing the properties.
	    this.constrainHeight = new Clutter.BindConstraint({ source: this._pageArea,
	                                                        coordinate: Clutter.BindCoordinate.HEIGHT });
	},
	byebye: function() {
		Main.overview.disconnect(this._connectionOv1)
		Main.overview.disconnect(this._connectionOv2)
		Main.overview.disconnect(this._connectionOv3)
		global.stage.disconnect(this._connectionGl1)
		if (this._stageKeyPressId != 0) global.stage.disconnect(this._stageKeyPressId)
		this._entry.disconnect(this._connectionSe1)
		this._text.disconnect(this._connectionCt1)
		this._text.disconnect(this._connectionCt2)
		this._text.disconnect(this._connectionCt3)
				
		global.focus_manager.remove_group(this._searchResults.actor)
		this._searchResults.actor.remove_actor(this._focusTrap)
		this._focusTrap.disconnect(this._connectionFc1)
		this._focusTrap = null
		
		let providers = this._searchSystem.getProviders()
		for (i in providers) this.removeSearchProvider(providers[i])
		
		this._searchSystem = null
		this._searchResults = null
		this._appDisplay = null
		this._workspacesDisplay = null
		this.constrainHeight = null
		
		this._inactiveIcon.destroy();	this._inactiveIcon = null
		this._activeIcon.destroy();		this._activeIcon = null
		this._pageArea.destroy();		this._pageArea = null
		this.actor.destroy();			this.actor = null
	}
})

/* Settings Guard */
let settingsGuard = null

function SettingsGuard() { this._init() }

SettingsGuard.prototype = {
	_init: function() {
		this.TermProvider = null
		
		this._termSettings = new Gio.Settings({ schema: TERM_SCHEMA })
		this._connectionTerm = this._termSettings.connect('changed::' + TERMNAME_KEY, this._listTerminals)	// see line 39
		this._listTerminals(this._termSettings)
		
		this._lockSettings = new Gio.Settings({ schema: LOCK_SCHEMA })
		this._connectionLock = this._lockSettings.connect('changed::' + NO_SHELL_KEY, Lang.bind(this, this._syncTermProvider))
		this._syncTermProvider()
	},
	
	// Terminals
	_listTerminals: function(_termSettings) {
		TERMS = new Array(_termSettings.get_string(TERMNAME_KEY))
		if (TERMS.indexOf("gnome-terminal") == -1)		TERMS.push("gnome-terminal")
		if (TERMS.indexOf(GLib.getenv('TERM')) == -1)	TERMS.push(GLib.getenv('TERM'))
		if (TERMS.indexOf("xfce4-terminal") == -1)		TERMS.push("xfce4-terminal")
		if (TERMS.indexOf("xterm") == -1)				TERMS.push("xterm")
	},
	
	// New Search Entry
	_insertNewSearchEntry: function() {
		let overview = Main.overview
		let group = overview._group
		let dash = overview._dash
		
		// Search Entry
		this._newSearchEntry = new ShellEntryWithTabKeyOK({	name: 'searchEntry', hint_text: S_("Type to search..."), track_hover: true, can_focus: true }, 													this.PathsCompleter)
		this._newSearchEntry.set_width(0.44 * Main.layoutManager.primaryMonitor.width)
		
		group.remove_actor(overview._searchEntry)
		overview._searchEntry = this._newSearchEntry
		group.add_actor(overview._searchEntry)
		
		// View Selector
		this.startSearch = ViewSelector.ViewSelector.prototype['startSearch']
		ViewSelector.ViewSelector.prototype['startSearch'] = function(event) {
			global.stage.set_key_focus(this._entry)		// or, after having 'esc'aping one time, the first character entered is written twice every time.
			this._text.event(event, true)
		}
		let providers = overview._viewSelector._searchSystem.getProviders()
		this._newViewSelector = new ViewSelectorWithNamedConnections(overview._searchEntry, dash.showAppsButton, providers)
		
		dash.actor.remove_constraint(overview._viewSelector.constrainHeight)
		group.remove_actor(dash.actor)
		group.remove_actor(overview._viewSelector.actor)
		overview._viewSelector = this._newViewSelector
		group.add_actor(overview._viewSelector.actor)
		group.add_actor(dash.actor)
		dash.actor.add_constraint(overview._viewSelector.constrainHeight)
		
		overview._relayout()
	},
	_removeNewSearchEntry: function() {
		let overview = Main.overview
		let group = overview._group
		let dash = overview._dash
		
		overview._group.remove_actor(overview._searchEntry)
		overview._searchEntry = new St.Entry({ name: 'searchEntry', hint_text: S_("Type to search..."), track_hover: true, can_focus: true })
		this._newSearchEntry.byebye()
		this._newSearchEntry = null
		overview._group.add_actor(overview._searchEntry)
		
		//ViewSelector.ViewSelector.prototype['startSearch'] = this.startSearch		// TODO : same problem now with a new viewSelector...
		
		dash.actor.remove_constraint(overview._viewSelector.constrainHeight)
		group.remove_actor(dash.actor)
		group.remove_actor(overview._viewSelector.actor)
		overview._viewSelector = new ViewSelector.ViewSelector(overview._searchEntry, dash.showAppsButton)
		this._newViewSelector.byebye()
		this._newViewSelector = null
		group.add_actor(overview._viewSelector.actor)
		group.add_actor(dash.actor)
		dash.actor.add_constraint(overview._viewSelector.constrainHeight)
		
		overview._relayout()
	},
	
	// PathsCompleter
	_enablePathsCompleter: function() {
		if (this.PathsCompleter) {
			global.logError("Why is there already a PathsCompleter!?")
			return }
		this.PathsCompleter = new Completor()
	},
	_disablePathsCompleter: function() {
		if (!this.PathsCompleter) {
			global.logError("No PathsCompleter to disable!?")
			return }
		this.PathsCompleter.byebye()
		this.PathsCompleter = null
	},
	
	// TermProvider
	_syncTermProvider: function() {
		if (!this._lockSettings.get_boolean(NO_SHELL_KEY)) {
			if (this.TermProvider) {
				global.logError("Why is there already a TermProvider!?")
				return }
			
			this._enablePathsCompleter()
			this._insertNewSearchEntry()
	
			this.TermProvider = new TermProvider(this.PathsCompleter)
			Main.overview.addSearchProvider(this.TermProvider)
		}
		else this._disableTermProvider()
	},
	_disableTermProvider: function() {
		if (!this.TermProvider) {
			global.logError("No TermProvider to disable!?")
			return }
		
		this._removeNewSearchEntry()
	
		Main.overview.removeSearchProvider(this.TermProvider)
		this.TermProvider.byebye()
		this.TermProvider = null
		
		this._disablePathsCompleter()
	},
	
	// Disable
	byebye: function() {
		this._disableTermProvider()
		
		this._lockSettings.disconnect(this._connectionLock); this._connectionLock = null; this._lockSettings = null
		this._termSettings.disconnect(this._connectionTerm); this._connectionTerm = null; this._termSettings = null
	}
}

/* appSearch things */
// This function is distributed in the appSearch extension
// https://extensions.gnome.org/extension/95/appsearch/
// author: LUV / Lukas Vacek <lucas.vacek@gmail.com>
// Distributed on extensions.gnome.org so you can use it as on GPL v. 2, licensed in GPL v. 3

// Ubuntu, Debian, Linux Mint 12, and PackageKit supported
// uses apt-cache show to search for packages
// If apt-cache is not available, packagekit used instead
// uses mint-make-cmd to install
let install = ""
function setInstallMode() {
	let test = GLib.spawn_command_line_sync("which apturl")
	if (!test[3]) {
		test = GLib.spawn_command_line_sync("which apt-cache")
		if (!test[3]) install = "apt"
	} else {
		test = GLib.spawn_command_line_sync("which gpk-install-package-name")
		if (!test[3]) test = GLib.spawn_command_line_sync("which pkcon")
		if (!test[3]) install = "pk"
	}
}

/* Extensions API */
function init(metadata) {
	MainFile.initTranslations(gettextDomain)
	lockFilePath = GLib.build_filenamev([metadata.path, "managed"])
//	setInstallMode()
}
function enable() {
	settingsGuard = new SettingsGuard()
	
	// The extension disables the <Alt>F2 keybinding until the user manually re-set it. If he does, a file is created in the extension's folder that "lock" the behaviour until the extension is removed. The file should be deleted at each manual unactivation of the extension, but that looks like impossible with the actual API.
	if (!GLib.file_test(lockFilePath, GLib.FileTest.EXISTS)) disableShellEntry()
}
function disable() {
	if (!GLib.file_test(lockFilePath, GLib.FileTest.EXISTS)) rEnableShellEntry()
	
	settingsGuard.byebye()
	settingsGuard = null
}
